#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import glob
import importlib.metadata
import logging
import os
import sys
from pathlib import Path
from typing import Dict

from carta_envoy.launchers import CartaLauncher
from pycapo import CapoConfig

# pylint: disable=C0301, E0401, W0622, W1203

"""
Setup and Launch CARTA via Workspaces
(modified from carta-valet)
"""

logger = logging.getLogger("carta_envoy")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def _get_settings(location: str) -> Dict:
    """
    Returns CARTA settings

    :param location: directory on which CARTA is to act
    :return: various CARTA settings
    """
    carta_settings = CapoConfig().settings("edu.nrao.carta")
    use_carta = CapoConfig().getboolean("edu.nrao.workspaces.carta.useCarta")
    notification_url = CapoConfig().settings("edu.nrao.workspaces.NotificationSettings").serviceUrl
    workflow_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl
    ui_url = CapoConfig().settings("edu.nrao.workspaces.UISettings").serviceUrl

    return {
        "timeout": carta_settings.timeoutInMinutes,
        "carta_path": carta_settings.pathToExecutable,
        "redis_host": carta_settings.redisHost,
        "redis_port": carta_settings.redisPort,
        "redis_password": carta_settings.redisPassword,
        "reverse_proxy": carta_settings.reverseProxyHost,
        "data_location": ensure_non_relative_path(location),
        "useCarta": use_carta,
        "notification_url": notification_url,
        "workflow_url": workflow_url,
        "ui_url": ui_url,
        "single_image": False,
    }


def get_carta_settings(location: str) -> Dict:
    """
    Accessor for protected function

    :param location: directory on which CARTA is to act
    :return: various CARTA settings
    """
    return _get_settings(location)


def ensure_non_relative_path(location: str) -> Path:
    """
    ensure the path for CARTA is non-relative

    :param location: string representation of the path
    :return: Path
    """
    if location.startswith("."):
        return Path(location).resolve()

    return Path(location)


def check_processing_dir(location: str) -> bool:
    return any(file.is_dir() and file.name == "working" for file in Path(location).glob("**/*"))


def ensure_correct_directory(settings: dict) -> str:
    """
    Ensure CARTA will start in the correct directory

    :param settings: settings dict
    :return:
    """

    directory = settings["data_location"]
    logger.info(f"Current working directory: {os.getcwd()}")

    files_count = len(glob.glob("*.fits"))
    is_processing_dir = check_processing_dir(directory)

    if os.getcwd() != Path(directory):
        os.chdir(directory)
    if files_count != 0:
        logger.info(f"FITS files found in location '{directory}', continuing...")
        return directory
    elif files_count == 0 and is_processing_dir:
        logger.info(f"Current directory is a processing directory, continuing...")
        return directory
    else:
        logger.info(f"No FITS files found in location '{directory}', checking other directories...")

        contained_dirs = glob.glob("*/", recursive=True)
        for dir in contained_dirs:
            if len(glob.glob(dir + "*.fits")) != 0:
                path = Path(dir)
                os.chdir(path)
                logger.info(f"Correct data location found: {path}. Changing location to match...")
                return str(path)


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Create command-line argument parser

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description="Launch CARTA via Workspaces",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=importlib.metadata.version("carta-envoy"),
    )
    parser.add_argument(
        "-d",
        "--directory",
        action="store",
        type=str,
        required=True,
        help="where to run CARTA",
    )
    parser.add_argument(
        "-n",
        "--notify",
        nargs=1,
        type=str,
        action="store",
        required=False,
        help="send user notification emails. (requires email address)",
    )
    parser.add_argument(
        "-r",
        "--ready",
        action="store",
        nargs=1,
        required=False,
        help="send 'carta ready' notification from Carta Envoy. Use in conjunction with -n|--notify",
    )
    parser.add_argument(
        "--parallel",
        action="store_true",
        required=False,
        help="allow CARTA envoy to run in parallel mode with WS productfetcher."
        " (Use when data must be retrieved for viewing)",
    )
    parser.add_argument(
        "--single",
        action="store_true",
        required=False,
        help="allow CARTA envoy to launch with specified image" " (Use when data must be retrieved for viewing)",
    )
    parser.add_argument(
        "-wf",
        "--workflow-request",
        nargs=1,
        type=int,
        required=False,
        help="Workflow request id when run as part of a WS workflow",
    )
    parser.add_argument(
        "-i",
        "--image",
        nargs=1,
        type=str,
        required=False,
        help="Name of specific image to display on launch",
    )
    return parser


def main():
    args = make_arg_parser().parse_args()
    settings = get_carta_settings(args.directory)

    if args.notify:
        settings["user_email"] = args.notify[0]
    if args.ready:
        settings["send_ready"] = str(args.ready[0]).lower()
    if args.workflow_request is not None:
        settings["wf_request_id"] = args.workflow_request[0]
    if args.image is not None:
        img_name = args.image[0]
        if "+" in img_name:
            # encode characters for url
            settings["image_name"] = img_name.replace("+", "%2B")
        else:
            settings["image_name"] = img_name
    if args.single:
        settings["single_image"] = True
    if not args.parallel:
        path = ensure_correct_directory(settings)
        settings["data_location"] = path

    carta = CartaLauncher(settings)
    carta.launch()
    carta.teardown()
    logger.info("CARTA session has completed.")
